# wave equation with cartesian PML and DG
# Joachim Schöberl

from ngsolve.comp import *
from ngsolve import *

from netgen.geom2d import SplineGeometry
geo = SplineGeometry()

pntx = [-1.5,-1,1,1.5]
pnty = [-1.5,-1,1,1.5]
pmlpnts = []
for yi in pnty:
    for xi in pntx:
        pmlpnts.append (geo.AddPoint(xi,yi))
#inner rect
geo.Append (["line", pmlpnts[5], pmlpnts[6]], leftdomain=1, rightdomain=3)
geo.Append (["line", pmlpnts[6], pmlpnts[10]], leftdomain=1, rightdomain=2)
geo.Append (["line", pmlpnts[10], pmlpnts[9]], leftdomain=1, rightdomain=3)
geo.Append (["line", pmlpnts[9], pmlpnts[5]], leftdomain=1, rightdomain=2)

# right x-pml
geo.Append (["line", pmlpnts[6], pmlpnts[7]], leftdomain=2, rightdomain=4)
geo.Append (["line", pmlpnts[7], pmlpnts[11]], leftdomain=2, rightdomain=0)
geo.Append (["line", pmlpnts[11], pmlpnts[10]], leftdomain=2, rightdomain=4)
# bottom y-pml
geo.Append (["line", pmlpnts[5], pmlpnts[1]], leftdomain=3, rightdomain=4)
geo.Append (["line", pmlpnts[1], pmlpnts[2]], leftdomain=3, rightdomain=0)
geo.Append (["line", pmlpnts[2], pmlpnts[6]], leftdomain=3, rightdomain=4)
# top y-pml
geo.Append (["line", pmlpnts[10], pmlpnts[14]], leftdomain=3, rightdomain=4)
geo.Append (["line", pmlpnts[14], pmlpnts[13]], leftdomain=3, rightdomain=0)
geo.Append (["line", pmlpnts[13], pmlpnts[9]], leftdomain=3, rightdomain=4)
# left x-pml
geo.Append (["line", pmlpnts[9], pmlpnts[8]], leftdomain=2, rightdomain=4)
geo.Append (["line", pmlpnts[8], pmlpnts[4]], leftdomain=2, rightdomain=0)
geo.Append (["line", pmlpnts[4], pmlpnts[5]], leftdomain=2, rightdomain=4)

# bottom right xy-pml
geo.Append (["line", pmlpnts[2], pmlpnts[3]], leftdomain=4, rightdomain=0)
geo.Append (["line", pmlpnts[3], pmlpnts[7]], leftdomain=4, rightdomain=0)
# top right xy-pml
geo.Append (["line", pmlpnts[11], pmlpnts[15]], leftdomain=4, rightdomain=0)
geo.Append (["line", pmlpnts[15], pmlpnts[14]], leftdomain=4, rightdomain=0)
# top left xy-pml
geo.Append (["line", pmlpnts[13], pmlpnts[12]], leftdomain=4, rightdomain=0)
geo.Append (["line", pmlpnts[12], pmlpnts[8]], leftdomain=4, rightdomain=0)
# bottom left xy-pml
geo.Append (["line", pmlpnts[4], pmlpnts[0]], leftdomain=4, rightdomain=0)
geo.Append (["line", pmlpnts[0], pmlpnts[1]], leftdomain=4, rightdomain=0)

geo.SetMaterial(1,"inner")
geo.SetMaterial(2,"pmlx")
geo.SetMaterial(3,"pmly")
geo.SetMaterial(4,"pmlxy")

mesh = Mesh( geo.GenerateMesh(maxh=0.1))
Draw (mesh)

k = 4


fes_u = VectorL2(mesh, order=k)
fes_p = L2(mesh, order=k+1) 
fes = FESpace( [fes_p,fes_p,fes_u, fes_u] )
p,phat, u,uhat = fes.TrialFunction()
q,qhat, v,vhat = fes.TestFunction()

n = specialcf.normal(2) 

B  = BilinearForm(fes)
B += SymbolicBFI ( grad(p)*v )
B += SymbolicBFI ( -0.5 * (p - p.Other()) * (n*v), element_boundary = True)

Bt  = BilinearForm(fes)
Bt += SymbolicBFI ( -grad(q)*u )
Bt += SymbolicBFI ( 0.5 * (q-q.Other()) * (n*u), element_boundary = True) 

ux = u[0]
uy = u[1]
vx = v[0]
vy = v[1]
uxhat = uhat[0]
uyhat = uhat[1]
vxhat = vhat[0]
vyhat = vhat[1]

# damping matrices
sigma = 10
dampp = BilinearForm(fes)
dampp += SymbolicBFI ( p*q, definedon=mesh.Materials("pmlx"))
dampp += SymbolicBFI ( p*q, definedon=mesh.Materials("pmly"))
dampp += SymbolicBFI ( 2*p*q +p*qhat-q*phat, definedon=mesh.Materials("pmlxy"))
dampu = BilinearForm(fes)
dampu += SymbolicBFI ( ux*vx,  definedon=mesh.Materials("pmlx"))
dampu += SymbolicBFI ( -uy*vy + uy*vyhat-vy*uyhat + uyhat*vyhat,  definedon=mesh.Materials("pmlx"))

dampu += SymbolicBFI ( -ux*vx + ux*vxhat-vx*uxhat + uxhat*vxhat,  definedon=mesh.Materials("pmly"))
dampu += SymbolicBFI ( uy*vy,  definedon=mesh.Materials("pmly"))

gfu = GridFunction(fes)
gfu.components[0].Set (  CoefficientFunction(exp ( -100 * ( (x-0.5)*(x-0.5)+(y+0.5)*(y+0.5) )) ) )

Draw(gfu.components[0], mesh, "p")
# Draw(gfu.components[1], mesh, "phat")
# Draw(gfu.components[2], mesh, "u")
# Draw(gfu.components[3], mesh, "uhat")

# For the time stepping
tau = 2e-3
tend = 10
t = 0

pdofs = fes.Range(0)
phatdofs = fes.Range(1)
udofs = fes.Range(2)
uhatdofs = fes.Range(3)

w = gfu.vec.CreateVector()
hv = gfu.vec.CreateVector()

with TaskManager():
    while t < tend:
        print (t)
        Bt.Apply (gfu.vec, w)
        dampp.Apply (gfu.vec, hv)
        w.data -= sigma * hv
        fes_p.SolveM (vec=w[pdofs])
        fes_p.SolveM (vec=w[phatdofs])
        gfu.vec.data += tau * w

        B.Apply (gfu.vec, w)
        dampu.Apply (gfu.vec, hv)
        w.data -= sigma * hv
        fes_u.SolveM (vec=w[udofs])
        fes_u.SolveM (vec=w[uhatdofs])
        gfu.vec.data += tau * w

        t += tau
        Redraw(False)



